@@ -11,14 +11,9 @@ use Carp;
 use Encode 'decode';
 use File::Temp 'tempfile';
 use FindBin '$Bin';
-use Gzip::Faster;
-use HTTP::Date;
-use JSON::Parse 'valid_json';
 use Test::Builder;
-use Unicode::UTF8 qw/decode_utf8 encode_utf8/;
-#use Path::Tiny;
 
-our $VERSION = '0.10';
+our $VERSION = '0.11';
 
 sub new
 {
@@ -51,7 +46,7 @@ sub on_off_msg
 	else {
 	    $msg .= "off";
 	}
-	$msg .= "testing of $type";
+	$msg .= " testing of $type";
 	my (undef, $file, $line) = caller ();
         $self->{tb}->note ("$file:$line: $msg");
     }
@@ -85,6 +80,15 @@ sub do_compression_test
     $switch = !! $switch;
     $self->on_off_msg ($switch, "compression");
     $self->{comp_test} = $switch;
+    if ($switch && ! $self->{_use_io_uncompress_gunzip}) {
+	eval "use Gzip::Faster;";
+	if ($@) {
+	    $self->{_use_io_uncompress_gunzip} = 1;
+	    if (! $self->{no_warn}) {
+		carp "Gzip::Faster is not installed, using IO::Uncompress::Gunzip";
+	    }
+	}
+    }
 }
 
 sub do_caching_test
@@ -93,15 +97,44 @@ sub do_caching_test
     $switch = !! $switch;
     $self->on_off_msg ($switch, "if-modified/last-modified response");
     $self->{cache_test} = $switch;
+    if ($switch) {
+	eval "use HTTP::Date;";
+	if ($@) {
+	    if (! $self->{no_warn}) {
+		carp "HTTP::Date is not installed, cannot do caching test";
+	    }
+	    $self->{cache_test} = undef;
+	}
+    }
 }
 
 sub expect_charset
 {
     my ($self, $charset) = @_;
+    eval "use Unicode::UTF8 qw/decode_utf8 encode_utf8/";
+    if ($@) {
+	Encode->import (qw/decode_utf8 encode_utf8/);
+	if (! $self->{no_warn} && ! $self->{_warned_unicode_utf8}) {
+	    carp "Unicode::UTF8 is not installed, using Encode";
+	    $self->{_warned_unicode_utf8} = 1;
+	}
+    }
     $self->note ("You have told me to expect a 'charset' value of '$charset'.");
     $self->{expected_charset} = $charset;
 }
 
+sub expect_mime_type
+{
+    my ($self, $mime_type) = @_;
+    if ($mime_type) {
+	$self->note ("You have told me to expect a mime type of '$mime_type'.");
+    }
+    else {
+	$self->note ("You have deleted the mime type.");
+    }
+    $self->{mime_type} = $mime_type;
+}
+
 sub set_verbosity
 {
     my ($self, $verbosity) = @_;
@@ -109,6 +142,13 @@ sub set_verbosity
     $self->note ("You have asked me to print messages as I work.");
 }
 
+sub set_no_warnings
+{
+    my ($self, $onoff) = @_;
+    $self->{no_warn} = !! $onoff;
+    $self->on_off_msg ($onoff, "warnings");
+}
+
 sub test_if_modified_since
 {
     my ($self, $last_modified) = @_;
@@ -116,6 +156,8 @@ sub test_if_modified_since
     my $saved = $ENV{HTTP_IF_MODIFIED_SINCE};
     $ENV{HTTP_IF_MODIFIED_SINCE} = $last_modified;
     $self->note ("Testing response with last modified time $last_modified");
+    my $saved_no_check_content = $self->{no_check_content};
+    $self->{no_check_content} = 1;
     # Copy the hash of options into a private copy, so that we can run
     # the thing again without overwriting our precious stuff.
     my $saved_run_options = $self->{run_options};
@@ -126,11 +168,15 @@ sub test_if_modified_since
     run_private ($self);
     $self->check_headers_private ($self);
     my $headers = $run_options{headers};
+    $self->do_test ($headers->{status} =~ /304/, "Got 304 status response");
+    my $body = $run_options{body};
+    $self->do_test (length ($body) == 0, "No body returned with 304 response");
     # Restore this to whatever it was, probably "undef".
     $ENV{HTTP_IF_MODIFIED_SINCE} = $saved;
     # Restore our precious stuff.
     $self->{run_options} = $saved_run_options;
     $self->{no_warn} = $saved_no_warn;
+    $self->{no_check_content} = $saved_no_check_content;
 }
 
 sub check_caching_private
@@ -240,6 +286,22 @@ sub setenv_private
     $ENV{$name} = $value;
 }
 
+sub encode_utf8_safe
+{
+    my ($self) = @_;
+    my $input = $self->{input};
+    eval "use Unicode::UTF8;";
+    if ($@) {
+	if (! $self->{no_warn} && ! $self->{_warned_unicode_utf8}) {
+	    carp "Unicode::UTF8 is not installed, using Encode";
+	    $self->{_warned_unicode_utf8} = 1;
+	}
+	# Encode::encode_utf8 uses prototypes so we have to hassle this up.
+	return Encode::encode_utf8 ($input);
+    }
+    return Unicode::UTF8::encode_utf8 ($input);
+}
+
 # Internal routine to run a CGI program.
 
 sub run_private
@@ -286,7 +348,7 @@ sub run_private
     if (defined $options->{input}) {
         $self->{input} = $options->{input};
 	if (utf8::is_utf8 ($self->{input})) {
-	    $self->{input} = encode_utf8 ($self->{input});
+	    $self->{input} = $self->encode_utf8_safe ();
 	}
         my $content_length = length ($self->{input});
         setenv_private ($self, 'CONTENT_LENGTH', $content_length);
@@ -377,7 +439,7 @@ my $field_content = qr/(?:$HTTP_TEXT)*|
                        )*
                       /x;
 
-my $http_token = qr/(?:$HTTP_TOKEN)+/;
+my $http_token = qr/(?:$HTTP_TOKEN+)/;
 
 # Check for a valid content type line.
 
@@ -386,53 +448,41 @@ sub check_content_line_private
     my ($self, $header, $verbose) = @_;
 
     my $expected_charset = $self->{expected_charset};
-    my $content_type_line;
 
     $self->note ("I am checking to see if the output contains a valid content type line.");
     my $content_type_ok;
-    if ($header =~ m!(Content-Type:\s*.*)!i) {
-        $self->pass_test ("There is a Content-Type header");
-        $content_type_line = $1;
-        if ($content_type_line =~ m!^Content-Type:(?:$HTTP_LWS)+
-                                        (?:$HTTP_TOKEN)+
-                                        /
-                                        (?:$HTTP_TOKEN)+
-                                   !xi) {
-            $self->pass_test ("The Content-Type header is well-formed");
-            if ($expected_charset) {
-                if ($content_type_line =~ /charset
-                                           =
-                                           (
-                                               $http_token|
-                                               $quoted_string
-                                           )/xi) {
-                    my $charset = $1;
-                    $charset =~ s/^"(.*)"$/$1/;
-                    if (lc $charset ne lc $expected_charset) {
-                        $self->fail_test ("You told me to expect a charset value of '$expected_charset', but the content-type line of the CGI executable, '$content_type_line', contains a charset parameter with the value '$charset'");
-                    }
-                    else {
-                        $content_type_ok = 1;
-                        $self->pass_test ("The charset '$charset' corresponds to the one you said to expect, '$expected_charset'");
-                    }
-                }
-                else {
-                    $self->fail_test ("You told me to expect a charset (character set) value of '$expected_charset', but the content-type line of the CGI executable, '$content_type_line', does not contain a valid 'charset' parameter");
-                }
-            }
-            else {
-                $content_type_ok = 1;
-                if ($verbose) {
-                    print "# I am not testing for the 'charset' parameter.\n";
-                }
-            }
-        }
-        else {
-            $self->fail_test ("The Content-Type line '$content_type_line' does not match the specification required.");
-        }
+    my $has_content_type = ($header =~ m!(Content-Type:\s*.*)!i);
+    my $content_type_line = $1;
+    $self->do_test ($has_content_type, "There is a Content-Type header");
+    if (! $has_content_type) {
+	return;
     }
-    else {
-        $self->fail_test ("There is no 'Content-Type' line in the output.");
+    my $lineok = ($content_type_line =~ m!^Content-Type:(?:$HTTP_LWS)+
+					  ($http_token/$http_token)
+					 !xi);
+    my $mime_type = $1;
+    $self->do_test ($lineok, "The Content-Type header is well-formed");
+    if (! $lineok) {
+	return;
+    }
+    if ($self->{mime_type}) {
+	$self->do_test ($mime_type eq $self->{mime_type},
+			"Got expected mime type $mime_type = $self->{mime_type}");
+    }
+    if ($expected_charset) {
+	my $has_charset = ($content_type_line =~ /charset
+						  =
+						  (
+						      $http_token|
+						      $quoted_string
+						  )/xi);
+	my $charset = $1;
+	$self->do_test ($has_charset, "Specifies a charset");
+	if ($has_charset) {
+	    $charset =~ s/^"(.*)"$/$1/;
+	    $self->do_test (lc $charset eq lc $expected_charset,
+			    "Got expected charset $charset = $expected_charset");
+	}
     }
 }
 
@@ -501,6 +551,32 @@ sub check_headers_private
     $self->{run_options}->{body} = $body;
 }
 
+# This is "safe" in the sense that it falls back to using
+# IO::Uncompress::Gunzip if it can't find Gzip::Faster. However, it
+# throws an exception if it fails, so it's not really "safe".
+
+sub gunzip_safe
+{
+    my ($self, $content) = @_;
+    my $out;
+    if ($self->{_use_io_uncompress_gunzip}) {
+	# gunzip_safe is called within an eval block. It's possible
+	# that the require might fail, but trying to fix these kinds
+	# of problems goes beyond the scope of this module.
+	eval "use IO::Uncompress::Gunzip;";
+	my $status = IO::Uncompress::Gunzip::gunzip (\$content, \$out);
+	if (! $status) {
+	    die "IO::Uncompress::Gunzip failed: $IO::Uncompress::Gunzip::GunzipError";
+	}
+    }
+    else {
+	# We have already loaded Gzip::Faster within
+	# do_compression_test.
+	$out = Gzip::Faster::gunzip ($content);
+    }
+    return $out;
+}
+
 sub check_compression_private
 {
     my ($self) = @_;
@@ -518,7 +594,7 @@ sub check_compression_private
         my $uncompressed;
         #printf "The length of the body is %d\n", length ($body);
 	eval {
-	    $uncompressed = gunzip $body;
+	    $uncompressed = $self->gunzip_safe ($body);
 	};
         if ($@) {
             $self->fail_test ("Output claims to be in gzip format but gunzip on the output failed with the error '$@'");
@@ -561,7 +637,9 @@ sub run
     }
     if (! $options) {
         $self->{run_options} = {};
-        carp "You have requested me to run a CGI executable with 'run' without specifying a hash reference to store the input, output, and error output. I can only run basic tests of correctness";
+	if (! $self->{no_warn}) {
+	    carp "You have requested me to run a CGI executable with 'run' without specifying a hash reference to store the input, output, and error output. I can only run basic tests of correctness";
+	}
     }
     else {
         $self->{run_options} = $options;
@@ -569,6 +647,28 @@ sub run
     if ($self->{verbose}) {
         print "# I am commencing the testing of CGI executable '$self->{cgi_executable}'.\n";
     }
+    if ($options->{html} && ! $self->{no_warn}) {
+	if ($self->{mime_type}) {
+	    if ($self->{mime_type} ne 'text/html') {
+		carp "If you want to test for HTML output, you should also specify a mime type 'text/html', but you have specified '$self->{mime_type}'";
+	    }
+	}
+	else {
+	    carp "If you want to check for html validity, you should also check the mime type is 'text/html' using expect_mime_type";
+	}
+    }
+    elsif ($options->{json} && ! $self->{no_warn}) {
+	my $mime_type = $self->{mime_type};
+	if ($mime_type) {
+	    if ($mime_type ne 'text/plain' && $mime_type ne 'application/json') {
+		carp "Your expected mime type of $mime_type is not valid for JSON";
+	    }
+	}
+	else {
+	    carp "There is no expected mime type, I suggest text/plain or application/json for JSON output";
+	}
+    }
+
 #    eval {
     run_private ($self);
     my $output = $self->{run_options}->{output};
@@ -690,11 +790,15 @@ sub set_html_validator
 {
     my ($self, $hvc) = @_;
     if (! $hvc) {
-	carp "Invalid value for validator";
+	if (! $self->{no_warn}) {
+	    carp "Invalid value for validator";
+	}
 	return;
     }
     if (! -x $hvc) {
-	carp "$hvc doesn't seem to be an executable program";
+	if (! $self->{no_warn}) {
+	    carp "$hvc doesn't seem to be an executable program";
+	}
     }
     $self->{html_validator} = $hvc;
 }
